#ifndef __FONT__
#define __FONT__

#include <X11/Xlib.h>

#ifdef DEBUG
#include <stdio.h>
#endif

typedef struct _tagFONTRECT
{
  int ascent;
  int descent;
  int width;
  int height;
} FONTRECT,*LPFONTRECT;

struct _tagFONT;

typedef int (*FONT_ADDREF)(struct _tagFONT *);
typedef int (*FONT_RELEASE)(struct _tagFONT *);
typedef int (*FONT_GETWIDTH)(struct _tagFONT *,const char *pText,int numChars);
typedef void (*FONT_GETCENTER)(struct _tagFONT *,const char *pText,int numChars,int width,int height,int *x,int *y);
typedef void (*FONT_GETEXTENTS)(struct _tagFONT *,const char *pText,int numChars,LPFONTRECT pRect);
typedef void (*FONT_RENDER)(struct _tagFONT *,Drawable wnd,int x,int y,const char *pText,int numChars);
typedef int (*FONT_GETID)(struct _tagFONT *);
typedef void (*FONT_SETCOLOR)(struct _tagFONT *,unsigned long uColor);
typedef void (*FONT_DESTROY)(struct _tagFONT *);

typedef struct _tagFONT
{
  FONT_ADDREF AddRef;		/* function pointers */
  FONT_RELEASE Release;
  FONT_GETWIDTH GetWidth;
  FONT_GETCENTER GetCenter;
  FONT_GETEXTENTS GetExtents;
  FONT_RENDER Render;
  FONT_GETID GetID;
  FONT_SETCOLOR SetColor;
  FONT_DESTROY Destroy;		/* DO NOT CALL THIS */

  unsigned int dwRefCount;	/* base class members */
  int ascent;
  int descent;
} FONT,*LPFONT;

typedef struct _tagXFONT
{
  FONT base;
  
  Display *m_pDisp;
  int m_iScreen;
  XFontStruct *m_pXfs;
  GC m_gc;
} XFONT,*LPXFONT;

LPFONT CreateXFont(Display *pDisp,int screen,Window mainWindow,const char *pName,int size);
#ifdef HAVE_XFT

#ifdef HAVE_FT2
#include <freetype/freetype.h>
#endif

#include <X11/Xft/Xft.h>

typedef struct _tagXFTFONT
{
  FONT base;

  Display *m_pDisp;
  int m_iScreen;
  XftFont *m_pFont;
  XftDraw *m_pDraw;
  XftColor m_color;
} XFTFONT,*LPXFTFONT;

LPFONT CreateXFTFont(Display *pDisp,int screen,Window mainWindow,const char *pName,int size);
#endif

#endif
