#include "config.h"
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#define XK_MISCELLANY
#include <X11/keysymdef.h>
#include <malloc.h>
#include <string.h>
#include "menu.h"
#include "pixmap.h"
#include "globals.h"
#include "atoms.h"
#include "defs.h"
#include "xhelp.h"

#define BORDER 2

static void FreeMenuItem(LPMENUITEM lpmi)
{
  /* do the callback, send destroy operation */
  MENUOPERATION mo;

  mo.cmd = MENU_OP_DESTROY;
  lpmi->callback(&mo,lpmi->param);

  /* check submenu */
  if (lpmi->pSubMenu)
  {
    DestroyMenu(lpmi->pSubMenu);
  }
}

static LPMENU FindMenu(LPMENU pMenu,Window wnd)
{
  LPMENUITEM pItem;
  LPMENU pRet;

  if (wnd == pMenu->wnd)
  {
    return pMenu;
  }
  /* failure, iterate through submenus now */
  if (pMenu->pList)
  {
    for(pItem = (LPMENUITEM)pMenu->pList->pHead;pItem;pItem = (LPMENUITEM)pItem->node.pNext)
    {
      if (pItem->pSubMenu)
      {
	if ((pRet = FindMenu(pItem->pSubMenu,wnd)))
	{
	  return pRet;
	}
      }
    }
  } 
  return NULL;
}

static LPMENUITEM HitTest(LPMENU pMenu,int x,int y)
{
  LPMENUITEM pItem;

  if (pMenu->pList)
  {
    for (pItem = (LPMENUITEM)pMenu->pList->pHead;pItem;pItem = (LPMENUITEM)pItem->node.pNext)
    {
      if ((x >= 0) && (x < pMenu->width) &&
	  ((pItem->y + pItem->height) > y) && (pItem->y <= y))
      {
	return pItem;
      }
    }
  }
  return NULL;
}

static void SetDrawFlags(LPMENU pMenu,LPMENUITEM pItem,unsigned int *drawFlags)
{
  *drawFlags = 0;

  if (pMenu->pSelected == pItem)
  {
    *drawFlags |= MENU_DF_SELECTED;
  }
  if (pItem->pSubMenu || (pItem->mit == MENUITEM_SUBMENU))
  {
    *drawFlags |= MENU_DF_SUBMENU;
  }
  if (pItem->mit == MENUITEM_SEPARATOR)
  {
    *drawFlags |= MENU_DF_SEPARATOR;
  }
}

static void RenderMenuItem(LPMENU pMenu,LPMENUITEM pItem)
{
  MENUOPERATION mo;
  /* set up the draw info struct */
  mo.cmd = MENU_OP_RENDER;
  mo.data.draw.drawable = pMenu->wnd;
  SetDrawFlags(pMenu,pItem,&mo.data.draw.drawFlags);
  
  mo.data.draw.x = pItem->x;
  mo.data.draw.y = pItem->y;
  mo.data.draw.width = pMenu->width - 2 * BORDER;
  mo.data.draw.height = pItem->height;
  /* now do the callback */
  pItem->callback(&mo,pItem->param);
}

static void RenderBorder(LPMENU pMenu)
{
  /* use gGc */
  
  /* draw outer border */
  XSetForeground(pDisp,gGc,PALETTE_LIGHT_GRAY);
  /* light gray on left side */
  XDrawLine(pDisp,pMenu->wnd,gGc,0,0,pMenu->width-1,0);
  XDrawLine(pDisp,pMenu->wnd,gGc,0,0,0,pMenu->height-1);
  /* black on right side */
  XSetForeground(pDisp,gGc,PALETTE_BLACK);
  XDrawLine(pDisp,pMenu->wnd,gGc,0,pMenu->height-1,pMenu->width-1,pMenu->height-1);
  XDrawLine(pDisp,pMenu->wnd,gGc,pMenu->width-1,0,pMenu->width-1,pMenu->height-1);
  /* now draw the inner border */

  /* white on the left side */
  XSetForeground(pDisp,gGc,PALETTE_WHITE);
  XDrawLine(pDisp,pMenu->wnd,gGc,1,1,pMenu->width-2,1);
  XDrawLine(pDisp,pMenu->wnd,gGc,1,1,1,pMenu->height-2);
  /* dark gray on the right side */
  XSetForeground(pDisp,gGc,PALETTE_DARK_GRAY);
  XDrawLine(pDisp,pMenu->wnd,gGc,1,pMenu->height-2,pMenu->width-2,pMenu->height-2);
  XDrawLine(pDisp,pMenu->wnd,gGc,pMenu->width-2,1,pMenu->width-2,pMenu->height-2);
}

static void Render(LPMENU pMenu)
{
  LPMENUITEM pItem;
  /* Render all the items */
  if (pMenu->pList)
  {
    for(pItem = (LPMENUITEM)pMenu->pList->pHead;pItem;pItem = (LPMENUITEM)pItem->node.pNext)
    {
      RenderMenuItem(pMenu,pItem);
    }
  }
  /* Now render the border */
  RenderBorder(pMenu);
}

static void Unselect(LPMENU pMenu)
{
  MENUOPERATION mo;

  if (pMenu->pSelected)
  {
    /* inform them */
    mo.cmd = MENU_OP_UNSELECTED;
    pMenu->pSelected->callback(&mo,pMenu->pSelected->param);
    /* hide the menu of this item */
    if (pMenu->pSelected->pSubMenu)
    {
      HideMenu(pMenu->pSelected->pSubMenu,0);
    }
    /* now clear */
    pMenu->pSelected = NULL;
  }
}

static void SelectItem(LPMENU pMenu,LPMENUITEM pItem)
{
  MENUOPERATION mo;
  int x,y;
  LPMENUITEM pOldSelection = pMenu->pSelected;
  /* unselect old menu */
  Unselect(pMenu);
  /* now update the new selection */
  pMenu->pSelected = pItem;
  /* tell them they're selected */
  mo.cmd = MENU_OP_SELECTED;
  pItem->callback(&mo,pItem->param);
  /* now redraw the two */
  RenderMenuItem(pMenu,pItem);
  if (pOldSelection)
  {
    RenderMenuItem(pMenu,pOldSelection);
  }
  /* check to see if we have to expand a menu */
  if (pItem->pSubMenu)
  {
    ClientToScreen(pMenu->wnd,pMenu->width - 2 * BORDER,pItem->y,&x,&y);
    ShowMenu(pItem->pSubMenu,0,x,y);
  }
}

static void OnMotion(LPMENU pMenu,XEvent *pEvent)
{
  LPMENUITEM pItem;

  pItem = HitTest(pMenu,pEvent->xmotion.x,pEvent->xmotion.y);
  if (pItem)
  {
    if (pMenu->pSelected != pItem)	/* a new match */
    {
      SelectItem(pMenu,pItem);
    }
    return;
  }
}

static int OnSelection(LPMENU pMenu,XEvent *pEvent)
{
  MENUOPERATION mo;
  LPMENUITEM pItem;

  pItem = HitTest(pMenu,pEvent->xbutton.x,pEvent->xbutton.y);
  if (!pItem)	/* not found, they missclicked */
  {
    return 1;
  }
  if (pItem->mit == MENUITEM_NORMAL) /* normal item, do the click */
  {
    /* now do the callback */
    mo.cmd = MENU_OP_CLICKED;
    pItem->callback(&mo,pItem->param);
    return 1;
  }
  return 0;
}

LPMENUITEM AddMenuItem(LPMENU pMenu,MENUCALLBACK callback,void *param,MENUITEMTYPE mit)
{
  LPMENUITEM lpmi;

  if (!pMenu->pList)
  {
    if (!(pMenu->pList = ListCreate((NODEFREEFUNC)FreeMenuItem,sizeof(LIST))))
    {
      return NULL;
    }
  }

  /* now add */
  lpmi = (LPMENUITEM)ListCreateNode(pMenu->pList,sizeof(MENUITEM));
  if (lpmi)
  {
    lpmi->callback = callback;
    lpmi->param = param;
    lpmi->mit = mit;
    lpmi->pSubMenu = NULL;
    /* now add to list */
    ListAddTail(pMenu->pList,(LPNODE)lpmi);
  }

  return lpmi;
}

LPMENU AddSubMenu(LPMENU pMenu,MENUCALLBACK callback,void *param)
{
  LPMENU pSub;
  LPMENUITEM pNewItem;
#ifdef __GNUC__
  MENUOPTIONS mops = { 
    .parent = pMenu->wnd,
    .background = pMenu->background_pixel
  };
#else
  MENUOPTIONS mops;
  mops.parent = pMenu->wnd;
  mops.background = pMenu->background_pixel;
#endif

  pSub = CreateMenu(&mops);
  pNewItem = AddMenuItem(pMenu,callback,param,MENUITEM_SUBMENU);
  pNewItem->pSubMenu = pSub;
  pSub->pParent = pMenu;

  return pSub;
}

int ProcessMenu(LPMENU pMenu,XEvent *pEvent,int *destroyed)
{
  LPMENU pOrig;
  
  *destroyed = 0;
 
  /* see if this item is destined for one of our windows */
  if (!pMenu->shown)
  {
    return 0;
  }
  
  pOrig = pMenu;
  if (!(pMenu=FindMenu(pMenu,pEvent->xany.window)))
  {
    if (pEvent->type == ButtonPress)
    {
      /* destroy myself */
      HideMenu(pOrig,1);
      *destroyed = 1;
      return 1;
    }
    return 0;
  }
 
  switch (pEvent->type)
  {
  case KeyPress:
    /* destroy on an escape */
    if (pEvent->xkey.keycode == XK_Escape ||
      pEvent->xkey.keycode == XK_Cancel ||
      pEvent->xkey.keycode == 0x09)
    {
      HideMenu(pMenu,1);
      *destroyed = 1;
    }
    break;
  case ButtonPress: 
    if (pEvent->xbutton.button == 1)
    {
      /* do stuff here, I pressed a button */
      if (OnSelection(pMenu,pEvent))
      {
	/* set destroyed flag */
	HideMenu(pMenu,1);
	*destroyed = 1;
      }
    }
    break;
  case Expose:
    if (pEvent->xexpose.count == 0)
    {
      Render(pMenu);
    }
    break;
  case MotionNotify:
    OnMotion(pMenu,pEvent);
    break;
  default:
    return 0;
  }
  return 1;
}

void ShowMenu(LPMENU pMenu,int type,int x,int y)
{
  unsigned int width = 0;
  unsigned int height = 0;
  LPMENUITEM pItem;
  MENUOPERATION mo;
  XWindowAttributes attr;
 
  /* calculate width and height */ 
  if (((pMenu->width == 0) && (pMenu->height == 0)) && pMenu->pList)
  {
    for (pItem = (LPMENUITEM)pMenu->pList->pHead;pItem;pItem = (LPMENUITEM)pItem->node.pNext)
    {
      /* setup operation to get size */
      mo.cmd = MENU_OP_GETSIZE;
      SetDrawFlags(pMenu,pItem,&mo.data.size.drawFlags);

      /* now call the function to get it */
      pItem->callback(&mo,pItem->param);
      
      pItem->x = BORDER;
      pItem->y = height + BORDER;
      pItem->width  = mo.data.size.width;
      pItem->height = mo.data.size.height; 

      /* check for max width */
      if (mo.data.size.width > width)
      {
	width = mo.data.size.width;
      }
      /* increment height */
      height += mo.data.size.height;
    }
    /* now set my width and height */
    pMenu->width  = width  + BORDER * 2;
    pMenu->height = height + BORDER * 2; 
  }

  /* get attributes for root window */
  XGetWindowAttributes(pDisp,root,&attr);
  /* move x,y around based on the type */
  if (type == MENU_SHOW_BOTTOM_LEFT)
  {
    y -= pMenu->height;
  }
  /* do range checks */
  if ((y + pMenu->height) > attr.height)
  {
    y = attr.height - pMenu->height;
  }
  if ((x + pMenu->width) > attr.width)
  {
    x = attr.width - pMenu->width;
  }

  /* now resize myself */
  XMoveResizeWindow(pDisp,pMenu->wnd,x,y,pMenu->width,pMenu->height);
  /* and display */
  XMapWindow(pDisp,pMenu->wnd);
  XRaiseWindow(pDisp,pMenu->wnd);
  pMenu->shown = 1;
  /* grab the pointer  */
  XGrabPointer(pDisp,pMenu->wnd,True,ButtonPressMask | PointerMotionMask,GrabModeAsync,GrabModeAsync,None,None,CurrentTime);
  /* XGrabKeyboard(pDisp,pMenu->wnd,True,GrabModeAsync,GrabModeAsync,CurrentTime);*/
}

void HideMenu(LPMENU pMenu,int hideParent)
{
  /* XWindowAttributes attr; */
  /* clear the selection */
  Unselect(pMenu);
  /* hide the window */
  XUnmapWindow(pDisp,pMenu->wnd);
  pMenu->shown = 0;
  /* Ungrab window */
  if (pMenu->pParent)
  {
    XGrabPointer(pDisp,pMenu->pParent->wnd,True,ButtonPressMask | PointerMotionMask,GrabModeAsync,GrabModeAsync,None,None,CurrentTime);
  }
  else
  {
    /* restore grabber like normal */
    XUngrabPointer(pDisp,CurrentTime);
  }
  /* XUngrabKeyboard(pDisp,CurrentTime); */
  /* close the parent */
  if (hideParent && (pMenu->pParent))
  {
    HideMenu(pMenu->pParent,hideParent);
  }
}

void DestroyMenu(LPMENU pMenu)
{
  XDestroyWindow(pDisp,pMenu->wnd);
  if (pMenu->pList)
  {
    ListDestroy(pMenu->pList);
  }
  free(pMenu);
}

LPMENU CreateMenu(LPMENUOPTIONS lpmop)
{
  XSetWindowAttributes att;
  MWMHints mwm;
  XSizeHints sizeHints;
  LPMENU pMenu = (LPMENU)malloc(sizeof(MENU));
  if (pMenu)
  {
    memset(pMenu,0,sizeof(MENU));
    
    pMenu->background_pixel = att.background_pixel = lpmop->background;
    att.event_mask =  ButtonPressMask | KeyPressMask | ExposureMask | PointerMotionMask;
    att.override_redirect = True;
    pMenu->wnd = XCreateWindow(pDisp,root,0,0,1,1,0,CopyFromParent,InputOutput,CopyFromParent,CWEventMask | CWBackPixel | CWOverrideRedirect,&att);
    
    /* borderless motif */
    memset(&mwm,0,sizeof(mwm));
    mwm.flags = MWM_HINTS_DECORATIONS;
    XChangeProperty(pDisp,pMenu->wnd,atom__MOTIF_WM_HINTS,atom__MOTIF_WM_HINTS,32,PropModeReplace,(unsigned char*)&mwm,sizeof(MWMHints)/4);

    sizeHints.flags = PPosition;
    XChangeProperty(pDisp,pMenu->wnd,XA_WM_NORMAL_HINTS,XA_WM_SIZE_HINTS,32,PropModeReplace,(unsigned char*)&sizeHints,sizeof(XSizeHints)/4);
  }
  return pMenu;
}
